/*
 * cryptengine.h
 *
 * Copyright (C) 2006 Jernimo Pellegrini
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to:
 *   The Free Software Foundation, Inc.,
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef CRYPTENGINE_H
#define CRYPTENGINE_H

#include <exception>
#include <stdexcept>
#include <new>
#include <string>
#include <list>

#include <boost/filesystem/path.hpp>
#include <boost/shared_ptr.hpp>

#include "key.h"
#include "bdata.h"

namespace apso {

/**
 * Wrappers over functions for cryptogtaphy.
 *
 * This class is an abstract "cryptographic engine". Real crypto engines are implemented as subclasses
 * of this one.
 */
class CryptEngine {
public:
	virtual char * random_bytes   (const unsigned size)                     throw(std::runtime_error, std::bad_alloc) = 0;
	virtual void   random_fill    (char *buffer, const unsigned size)       throw(std::runtime_error)                 = 0;

 	virtual char * sym_enc        (const char *text,    const char *key, unsigned *size, const unsigned keysize) throw(std::runtime_error, std::bad_alloc) = 0;
	virtual char * sym_dec        (const char *enctext, const char *key, unsigned *size, const unsigned keysize) throw(std::runtime_error, std::bad_alloc) = 0;
	virtual char * asym_enc       (const char *text,    const char *key, unsigned *size, const unsigned keysize) throw(std::runtime_error, std::bad_alloc) = 0;
	virtual char * asym_dec       (const char *text,    const char *key, unsigned *size, const unsigned keysize) throw(std::runtime_error, std::bad_alloc) = 0;
	virtual char * hash           (const char *text,    unsigned *size) throw(std::runtime_error, std::bad_alloc) = 0;
	virtual char * encode         (const char *text,    unsigned *size) throw(std::runtime_error, std::bad_alloc) = 0;
	virtual char * decode         (const char *text,    unsigned *size) throw(std::runtime_error, std::bad_alloc) = 0;
	virtual char * encode16       (      char *text,    unsigned *size) throw(std::runtime_error, std::bad_alloc) = 0;

        virtual bdata_ptr   sym_enc   (const bdata_ptr text, const Key& key) throw(std::runtime_error, std::bad_alloc) = 0;
        virtual bdata_ptr   sym_dec   (const bdata_ptr text, const Key& key) throw(std::runtime_error, std::bad_alloc) = 0;
        virtual bdata_ptr   asym_enc  (const bdata_ptr text, const Key& key) throw(std::runtime_error, std::bad_alloc) = 0;
        virtual bdata_ptr   asym_dec  (const bdata_ptr text, const Key& key) throw(std::runtime_error, std::bad_alloc) = 0;

	char * sym_enc  (const char *text,    Key& key,  unsigned *size) throw (std::runtime_error, std::bad_alloc);
	char * sym_dec  (const char *enctext, Key& key,  unsigned *size) throw (std::runtime_error, std::bad_alloc);
	char * asym_enc (const char *text,    Key& pubk, unsigned *size) throw (std::runtime_error, std::bad_alloc);
	char * asym_dec (const char *text,    Key& prik, unsigned *size) throw (std::runtime_error, std::bad_alloc);

        std::string sym_enc  (const std::string& text,    Key& key,  unsigned *size) throw (std::runtime_error, std::bad_alloc);
        std::string sym_dec  (const std::string& enctext, Key& key,  unsigned *size) throw (std::runtime_error, std::bad_alloc);
        std::string asym_enc (const std::string& text,    Key& pubk, unsigned *size) throw (std::runtime_error, std::bad_alloc);
        std::string asym_dec (const std::string& text,    Key& prik, unsigned *size) throw (std::runtime_error, std::bad_alloc);

	CryptEngine();
	virtual ~CryptEngine();
};

typedef boost::shared_ptr<CryptEngine> CryptEngine_ptr;

}

#endif

